﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using RDFMatViewCalculation.Logic;
using System.IO;
using System.Runtime.Serialization.Formatters.Binary;

namespace RDFMatViewCalculation
{
    class Program
    {
        static void Main(string[] args)
        {
            if (args != null && args.Length > 0 && args[0].Equals("-load"))
            {
                System.Diagnostics.Stopwatch watch = System.Diagnostics.Stopwatch.StartNew();
                LoadViews();
                watch.Stop();
                long elapsedMs = watch.ElapsedMilliseconds;
                Logging.Logger.log("Execution took " + elapsedMs + " ms");
                Environment.Exit(0);
            }

            //double Completeness = 0.75;
            //TestConnection tc = new TestConnection("http://schema.lubm.org/", "http://localhost:8890/sparql", "http://lubm.org/", "http://localhost:8890/sparql");
            //TestConnection tc = new TestConnection("http://schema.dbpedia.org/person/", "http://localhost:8890/sparql", "http://dbpedia.org/infobox", "http://164.15.78.105:8890/sparql");
            //TestConnection tc = new TestConnection("http://ssb.org/schemas/", "http://164.15.78.105:8080/openrdf-sesame/repositories/ssbCompleteSchema", "http://ssb.org/data/", "http://localhost:8890/sparql");
            TestConnection tc = new TestConnection("http://ssb.org/schemas/", "http://localhost:8890/sparql", "http://ssb.org/data/", "http://localhost:8890/sparql");
            //TestConnection tc = new TestConnection("http://ssb.org/schemas/", "http://localhost:8890/sparql", "http://ssb.org/data/", "http://localhost:8890/sparql");
            List<Hierarchy> allHiers = new List<Hierarchy>();
            List<List<HierarchyLevel>> dimHierarchies = new List<List<HierarchyLevel>>();
            //Console.WriteLine(tc.GetObservationCount());
            List<string> dims = tc.GetDimensionNames();
            foreach (string dimName in dims)
            {
                string hierName = tc.GetDimensionHierarchy(dimName);
                Console.WriteLine("Hierarchy Name: " + hierName);
                List<HierarchyStep> hierSteps = tc.GetHierarchySteps(hierName);
                List<HierarchyLevel> hierLvls = tc.GetHierarchyLevels(hierName, dimName);
                dimHierarchies.Add(hierLvls);
                Hierarchy hier = new Hierarchy() { DimensionName = dimName, Name = hierName, HierarchyLevels = hierLvls, HierarchySteps = hierSteps };
                allHiers.Add(hier);
                StringBuilder sb = new StringBuilder();
                foreach (Hierarchy item in allHiers)
                {
                    sb.AppendLine(string.Format("Hierarchy name: {0}, number of levels: {1}, number of steps: {2}",
                        item.Name, item.HierarchyLevels.Count, item.HierarchySteps.Count));
                }
                Logging.Logger.log(sb.ToString());
            }

            ////all permutations
            List<List<string>> flatLattice = tc.GetDimensionCombinations(dims);
            StringBuilder sb1 = new StringBuilder();
            foreach (List<string> item in flatLattice)
            {
                sb1.AppendLine(string.Format("Dimension combinations: {0}", string.Join(", ", item)));
            }
            Logging.Logger.log(sb1.ToString());

            ////all dimensions with hierarchies
            List<CuboidNode> allCuboidNodes = new List<CuboidNode>();
            foreach (List<string> dimCombinations in flatLattice)
            {
                List<List<HierarchyLevel>> parts = new List<List<HierarchyLevel>>();
                foreach (string dim in dimCombinations)
                {
                    parts.Add(dimHierarchies.SelectMany(c => c).Where(c => c.DimensionName.Equals(dim)).ToList());//GetDimensionParts(dim));
                }
                List<CuboidNode> bothNodes = tc.GetPermutations(parts);
                allCuboidNodes.AddRange(bothNodes);
            }
            sb1 = new StringBuilder();
            foreach (CuboidNode item in allCuboidNodes)
            {
                sb1.AppendLine(item.ToString());
            }
            Logging.Logger.log(sb1.ToString());

            //// now for every cuboid build a query to build a view and its size
            List<View> allViews = new List<View>();
            int allTriplesCount = tc.GetAllTriplesCount();
            int i = 1;
            Logging.Logger.log("Total Cuboid queries " + allCuboidNodes.Count);
            foreach (CuboidNode item in allCuboidNodes)
            {
                List<ViewPath> viewPathsList = tc.GetViewPaths(item, allHiers);
                //if (viewPathsList.Count > 1)
                //    Console.WriteLine("multiple paths");
                foreach (ViewPath vp in viewPathsList)
                {
                    string sparql = tc.BuildCountSPARQLfromPathList(item, vp.Dimensions.SelectMany(c => c.Steps).ToList(), allHiers);
                    Logging.Logger.log("Cuboid node query from PATH " + i.ToString() + ": " + sparql);
                    int size = tc.GetViewSize(sparql);
                    if(size == 0)
                        Logging.Logger.log("Cuboid node query from PATH " + i.ToString() + ": " + sparql + " size 0 #######################");
                    vp.Cost = size;
                }
                int maxSize = viewPathsList.Select(c => c.Cost).Max();
                ViewPath cvp = viewPathsList.Where(c => !c.IsPartial).FirstOrDefault();
                if (cvp != null) //// if the complete path exists - maxSize is the cost of complete size
                    maxSize = viewPathsList.Where(c => !c.IsPartial).Select(c => c.Cost).Max();
                else //// else the whole view is the sum of all parts
                    maxSize = viewPathsList.Select(c => c.Cost).Sum();
                foreach (ViewPath vp in viewPathsList)
                {
                    vp.Completeness = ((double)vp.Cost) / maxSize;
                }
                
                i++;

                double completeness = 1;

                //// this needs changing - the normal sparql query should take into account that there are several path that may compliment each other
                string sparql_O = tc.BuildSPARQLfromCuboid_V3(item, viewPathsList, allHiers); //tc.BuildSPARQLfromCuboid(item, allHiers);
                Logging.Logger.logSeparately("Cuboid node query total " + i.ToString() + ": " + sparql_O);
                int size_O = tc.GetViewSize(sparql_O);
                if(size_O == 0)
                    Logging.Logger.log("Cuboid node query from PATH " + i.ToString() + ": " + sparql_O + " size 0 #######################");
                View v = new View() { Node = item, Size = size_O, Cost = allTriplesCount, Completeness = completeness, Paths = viewPathsList , FactBaseSize = maxSize}; ////Cost = maxSize, - no, it is fact number size, not cost
                //tc.SetViewCompleteness(v, allHiers);
                allViews.Add(v);

                System.Threading.Thread.Sleep(500);
            }

            /////save the thing to a file.
            FileStream fs = new FileStream("views.dat", FileMode.Create);
            BinaryFormatter bf = new BinaryFormatter();
            bf.Serialize(fs, allViews);
            fs.Close();
            fs.Dispose();

            fs = new FileStream("hierarchies.dat", FileMode.Create);
            bf = new BinaryFormatter();
            bf.Serialize(fs, allHiers);
            fs.Close();
            fs.Dispose();

            fs = new FileStream("cuboid.dat", FileMode.Create);
            bf = new BinaryFormatter();
            bf.Serialize(fs, allCuboidNodes);
            fs.Close();
            fs.Dispose();

            ////// if one viewpath has a completeness less than 1.0 - lower level views should be "punished" - their completeness should be reduced
            //List<ViewPath> allIncompleteViewPaths = 
            //    allViews.SelectMany(c => c.Paths).Where(d => (d.Completeness < 1 && d.Dimensions.SelectMany(k => k.Steps).Where(e => e.IsOptional).Count() == 0 && d.Dimensions.Count == 1)).ToList();

            //List<ViewPath> neededIncompleteViewPath = new List<ViewPath>();
            //List<ViewPath> allPossCompleteViewPaths =
            //    allViews.SelectMany(c => c.Paths).Where(d => (d.Completeness == 1 && d.Dimensions.SelectMany(k => k.Steps).Where(e => e.IsOptional).Count() == 0 && d.Dimensions.Count == 1)).ToList();
            //foreach (ViewPath vpInc in allIncompleteViewPaths)
            //{
            //    List<HierarchyStep> incompletePathSteps = vpInc.Dimensions.SelectMany(c => c.Steps).ToList();
            //    foreach (ViewPath vpC in allPossCompleteViewPaths)
            //    {
            //        List<HierarchyStep> completePathSteps = vpC.Dimensions.SelectMany(c => c.Steps).ToList();
            //        if (incompletePathSteps.Intersect(completePathSteps).Count() == completePathSteps.Count && completePathSteps.Except(incompletePathSteps).Count() == 0)
            //        {
            //            if (!neededIncompleteViewPath.Contains(vpC))
            //            {
            //                vpC.Completeness = vpC.Completeness * vpInc.Completeness;
            //                neededIncompleteViewPath.Add(vpC);
            //            }
            //        }
            //    }
            //}

            //foreach (ViewPath vp in neededIncompleteViewPath)
            //{
            //    List<ViewPath> allCompleteViewPaths = allViews.SelectMany(c => c.Paths).Where(d => (d.Completeness == 1 && d.Dimensions.SelectMany(k => k.Steps).Where(e => e.IsOptional).Count() == 0)).ToList();
            //    List<HierarchyStep> incompletePathSteps = vp.Dimensions.SelectMany(c => c.Steps).ToList();
            //    string hierName = vp.Dimensions[0].HierarchyName;
            //    foreach (ViewPath vpC in allCompleteViewPaths)
            //    {
            //        List<HierarchyStep> completePathSteps = vpC.Dimensions.Where(c => c.HierarchyName.Equals(hierName)).SelectMany(c => c.Steps).ToList();
            //        if (incompletePathSteps.Intersect(completePathSteps).Count() > 0 && completePathSteps.Except(incompletePathSteps).Count() == 0)
            //        {
            //            vpC.Completeness = vpC.Completeness * vp.Completeness;
            //            //break;
            //        }
            //    }
            //    //Console.WriteLine("found" + vp.ToString());
            //}
            //List<ViewPath> allIncompleteViewPaths2 = allViews.SelectMany(c => c.Paths).Where(d => (d.Completeness < 1)).ToList();
            //Console.WriteLine("found" + allIncompleteViewPaths2.ToString());

            //// if the upper level view does not include all view paths of a previous view - it is not complete??? 
            //// or the path is not marked as complete



            //// calculate which one can be used to calculate which one.
            int numOfViewsToMat = 10;
            List<View> materializedViews = new List<View>();
            List<View> copyOfAllViews = allViews.Select(c => c).ToList();
            for (int j = 0; j < numOfViewsToMat; j++)
            {
                if (copyOfAllViews.Count < 1)
                    break;
                View maxBenView = null;
                int maxBenefit = 0;
                foreach (View v in copyOfAllViews)
                {
                    //int ben = tc.GetViewBenefit(v, copyOfAllViews, allHiers);//, materializedViews);
                    //if (ben > maxBenefit && v.Completeness > Completeness) /////&& v.Completeness > 0.9 - right now hard coded saying that completeness should be higher than a specific level
                    //{
                    //    maxBenView = v;
                    //    maxBenefit = ben;
                    //}
                    //if (v.Node.NodeParts.Where(c => !c.IsViewable).FirstOrDefault() == null)
                    //{
                        int ben = tc.GetViewBenefit_V3(v, copyOfAllViews, allHiers);//tc.GetViewBenefit_V2(v, copyOfAllViews, allHiers, Completeness);
                        if (ben > maxBenefit)
                        {
                            maxBenView = v;
                            maxBenefit = ben;
                        }
                    //}
                }
                
                //// if view is null (not used to answer other views) - select one with max benefit among the current views
                if (maxBenView == null)
                {
                    int MaxBen = copyOfAllViews.Select(c => (c.Cost - c.Size)).Max();
                    maxBenView = copyOfAllViews.Where(c => ((c.Cost - c.Size) == MaxBen)).FirstOrDefault();
                }

                //double maxCompleteness = maxBenView.Paths.Select(c => c.Completeness).Max();
                //ViewPath selectedVP = maxBenView.Paths.Where(c => c.Completeness == maxCompleteness).FirstOrDefault();
                //maxBenView.SelectedPath = selectedVP;
                Logging.Logger.log("View with max benefit: " + maxBenView.ToString() + ", benefit: " + maxBenefit.ToString());

                //Logging.Logger.log(GetViewBenefitForLogging(maxBenView, copyOfAllViews));//// comments about the view

                copyOfAllViews.Remove(maxBenView);//// remove this view from list to get ythe next view
                materializedViews.Add(maxBenView);//// add it to already materialized views
                //// update all remaining view costs taking into account materialized views
                foreach (View testView in copyOfAllViews)
                {
                    View matView = tc.GetBestMatViewForView_V3(testView, materializedViews, allHiers);
                    //GetBestMatViewForView_V2(testView, materializedViews, allHiers, Completeness);//// get the best matView for the view in question
                    if (matView != null && testView.Cost > matView.Size) //// if such view exists - change the cost 
                        testView.Cost = matView.Size;
                }
            }
        }

        private static void LoadViews()
        {
            TestConnection tc = new TestConnection("http://ssb.org/schemas/", "http://164.15.78.105:8080/openrdf-sesame/repositories/ssbCompleteSchema", "http://ssb.org/data/", "http://localhost:8890/sparql");
            List<View> allViews = new List<View>();
            List<Hierarchy> allHiers = new List<Hierarchy>();
            long totalBenefit = 0;

            FileStream fs = new FileStream("views.dat", FileMode.Open);
            BinaryFormatter bf = new BinaryFormatter();
            allViews = (List<View>)bf.Deserialize(fs);
            fs.Close();
            fs.Dispose();

            fs = new FileStream("hierarchies.dat", FileMode.Open);
            bf = new BinaryFormatter();
            allHiers = (List<Hierarchy>)bf.Deserialize(fs);
            fs.Close();
            fs.Dispose();

            int numOfViewsToMat = 50;
            List<View> materializedViews = new List<View>();
            List<View> copyOfAllViews = allViews.Select(c => c).ToList();
            for (int j = 0; j < numOfViewsToMat; j++)
            {
                if (copyOfAllViews.Count < 1)
                    break;
                View maxBenView = null;
                //int maxBenefit = 0;
                double maxBenefit = 0;
                foreach (View v in copyOfAllViews)
                {
                    //int ben = tc.GetViewBenefit(v, copyOfAllViews, allHiers);//, materializedViews);
                    //if (ben > maxBenefit && v.Completeness > Completeness) /////&& v.Completeness > 0.9 - right now hard coded saying that completeness should be higher than a specific level
                    //{
                    //    maxBenView = v;
                    //    maxBenefit = ben;
                    //}
                    //if (v.Node.NodeParts.Where(c => !c.IsViewable).FirstOrDefault() == null)
                    //{
                        //int ben = tc.GetViewBenefit_V3(v, copyOfAllViews, allHiers);//tc.GetViewBenefit_V2(v, copyOfAllViews, allHiers, Completeness);
                    double ben = tc.GetViewBenefit_V4(v, copyOfAllViews, allHiers);//tc.GetViewBenefit_V2(v, copyOfAllViews, allHiers, Completeness);
                        if (ben > maxBenefit)
                        {
                            maxBenView = v;
                            maxBenefit = ben;
                        }
                    //}
                }
                totalBenefit += (long)maxBenefit;

                //// if view is null (not used to answer other views) - select one with max benefit among the current views
                if (maxBenView == null)
                {
                    int MaxBen = copyOfAllViews.Select(c => (c.Cost - c.Size)).Max();
                    totalBenefit += MaxBen;
                    maxBenView = copyOfAllViews.Where(c => ((c.Cost - c.Size) == MaxBen)).FirstOrDefault();
                }

                //double maxCompleteness = maxBenView.Paths.Select(c => c.Completeness).Max();
                //ViewPath selectedVP = maxBenView.Paths.Where(c => c.Completeness == maxCompleteness).FirstOrDefault();
                //maxBenView.SelectedPath = selectedVP;
                Logging.Logger.log("View with max benefit: " + maxBenView.ToString() + ", benefit: " + maxBenefit.ToString());

                //Logging.Logger.log(GetViewBenefitForLogging(maxBenView, copyOfAllViews));//// comments about the view

                copyOfAllViews.Remove(maxBenView);//// remove this view from list to get ythe next view
                materializedViews.Add(maxBenView);//// add it to already materialized views
                //// update all remaining view costs taking into account materialized views
                foreach (View testView in copyOfAllViews)
                {
                    View matView = tc.GetBestMatViewForView_V3(testView, materializedViews, allHiers);
                    //GetBestMatViewForView_V2(testView, materializedViews, allHiers, Completeness);//// get the best matView for the view in question
                    if (matView != null && testView.Cost > matView.Size) //// if such view exists - change the cost 
                        testView.Cost = matView.Size;
                }
                
                ///// now calculate total time, total benefit and total space
                long totalSpace = materializedViews.Select(c => c.Size).Sum();
                long totalTime = copyOfAllViews.Select(c => c.Cost).Sum();
                Logging.Logger.logSeparately(string.Format("number of views: {0},{1},{2},{3}", materializedViews.Count, totalBenefit, totalTime, totalSpace));
            }
        }
    }
}
