﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using VDS.RDF.Parsing;
using VDS.RDF.Query;
using System.Data;
using VDS.RDF;
using RDFMatViewCalculation.Logic;

namespace RDFMatViewCalculation
{
    public class TestConnection
    {
        private string schemaGraphIRI;
        private string dataGraphIRI;
        private Uri sparqlEndpointUri;
        private Uri schemaSparqlEndpointUri;
        private static List<MeasureInfo> measures;

        public TestConnection(string schemaGraphURI, string schemaEndpointURI, string dataGraphURI, string dataEndpointURI)
        {
            schemaGraphIRI = schemaGraphURI;
            schemaSparqlEndpointUri = new Uri(schemaEndpointURI);
            dataGraphIRI = dataGraphURI;
            sparqlEndpointUri = new Uri(dataEndpointURI); ////"http://localhost:8890/sparql"
        }

        private static string _observationCountQuery;
        
        public string GetObservationCountQuery()
        {
            if (string.IsNullOrEmpty(_observationCountQuery))
            {
                List<string> obsTypes = new List<string>();
                SparqlQuery mainQuery = null;
                SparqlQueryParser parser = new SparqlQueryParser();
                string cmdSparql = SparqlCommands.CmdPrefixes + string.Format(SparqlCommands.CmdObservationDatatype, schemaGraphIRI);
                mainQuery = parser.ParseFromString(cmdSparql);
                mainQuery.Timeout = 600 * 1000;

                SparqlRemoteEndpoint sre = new SparqlRemoteEndpoint(schemaSparqlEndpointUri);
                sre.Timeout = 600 * 1000;
                RemoteQueryProcessor rqp = new RemoteQueryProcessor(sre);
                object obj = rqp.ProcessQuery(mainQuery);
                string str = mainQuery.ToString();
                Console.WriteLine(str);
                if (obj is SparqlResultSet)
                {
                    SparqlResultSet rSet = (SparqlResultSet)obj;
                    DataTable resultsTable = ConvertResultToDataTable(rSet);
                    foreach (DataRow item in resultsTable.Rows)
                    {
                        string type = item["type"].ToString();
                        if (!obsTypes.Contains(type)) obsTypes.Add(type);
                    }
                }
                StringBuilder sb = new StringBuilder();
                foreach (string tp in obsTypes)
                {
                    if (sb.ToString().Length > 3)
                        sb.Append(" || ");
                    sb.Append(string.Format("?type=<{0}>", tp));
                }
                _observationCountQuery = string.Format(@"SELECT (COUNT(*) AS ?count) FROM <{0}> WHERE {{
            SELECT DISTINCT ?obs
            WHERE {{
                ?obs  <http://www.w3.org/1999/02/22-rdf-syntax-ns#type> ?type .
                FILTER({1})
            }}
}}", dataGraphIRI, sb.ToString());
            }
            return _observationCountQuery;
        }

        public List<string> GetDimensionNames()
        {
            List<string> dimensionNames = new List<string>();
            SparqlQuery mainQuery = null;
            SparqlQueryParser parser = new SparqlQueryParser();
            mainQuery = parser.ParseFromString(SparqlCommands.CmdPrefixes + string.Format(SparqlCommands.CmdDimensionNames, schemaGraphIRI));
            mainQuery.Timeout = 7200 * 1000;

            SparqlRemoteEndpoint sre = new SparqlRemoteEndpoint(schemaSparqlEndpointUri);
            sre.Timeout = 7200 * 1000;
            RemoteQueryProcessor rqp = new RemoteQueryProcessor(sre);
            object obj = rqp.ProcessQuery(mainQuery);
            string str = mainQuery.ToString();
            Console.WriteLine(str);
            if (obj is SparqlResultSet)
            {
                SparqlResultSet rSet = (SparqlResultSet)obj;
                DataTable resultsTable = ConvertResultToDataTable(rSet);
                foreach (DataRow item in resultsTable.Rows)
                {
                    string dimName = item["dim"].ToString();
                    if (!dimensionNames.Contains(dimName))
                        dimensionNames.Add(dimName);
                }
            }
            return dimensionNames;
        }

        public string GetDimensionHierarchy(string dimName)
        {
            List<string> dimensionNames = new List<string>();
            SparqlQuery mainQuery = null;
            SparqlQueryParser parser = new SparqlQueryParser();
            string cmdSparql = SparqlCommands.CmdPrefixes + string.Format(SparqlCommands.CmdDimensionHierarchy, schemaGraphIRI, dimName);
            mainQuery = parser.ParseFromString(cmdSparql);
            mainQuery.Timeout = 7200 * 1000;

            SparqlRemoteEndpoint sre = new SparqlRemoteEndpoint(schemaSparqlEndpointUri);
            sre.Timeout = 7200 * 1000;
            RemoteQueryProcessor rqp = new RemoteQueryProcessor(sre);
            object obj = rqp.ProcessQuery(mainQuery);
            string str = mainQuery.ToString();
            Console.WriteLine(str);
            if (obj is SparqlResultSet)
            {
                SparqlResultSet rSet = (SparqlResultSet)obj;
                DataTable resultsTable = ConvertResultToDataTable(rSet);
                foreach (DataRow item in resultsTable.Rows)
                {
                    string hierName = item["hier"].ToString();
                    return hierName;
                }
            }
            return null;
        }

        public List<HierarchyStep> GetHierarchySteps(string hierName)
        {
            List<HierarchyStep> hierSteps = new List<HierarchyStep>();
            SparqlQuery mainQuery = null;
            SparqlQueryParser parser = new SparqlQueryParser();
            string cmdSparql = SparqlCommands.CmdPrefixes + string.Format(SparqlCommands.CmdHierarchySteps, schemaGraphIRI, hierName);
            mainQuery = parser.ParseFromString(cmdSparql);
            mainQuery.Timeout = 7200 * 1000;

            SparqlRemoteEndpoint sre = new SparqlRemoteEndpoint(schemaSparqlEndpointUri);
            sre.Timeout = 7200 * 1000;
            RemoteQueryProcessor rqp = new RemoteQueryProcessor(sre);
            object obj = rqp.ProcessQuery(mainQuery);
            string str = mainQuery.ToString();
            Console.WriteLine(str);
            if (obj is SparqlResultSet)
            {
                SparqlResultSet rSet = (SparqlResultSet)obj;
                DataTable resultsTable = ConvertResultToDataTable(rSet);
                foreach (DataRow item in resultsTable.Rows)
                {
                    string chLevelName = item["cl"].ToString();
                    string parLevelName = item["pl"].ToString();
                    string predicName = item["pred"].ToString();
                    string direction = item["dir"].ToString();
                    string stepName = item["s"].ToString();
                    string optional = "0";
                    if (item["mand"] != null && !string.IsNullOrEmpty(item["mand"].ToString()))
                    {
                        optional = item["mand"].ToString();
                    }
                    string partial = "0";
                    if (item["part"] != null && !string.IsNullOrEmpty(item["part"].ToString()))
                    {
                        partial = item["part"].ToString();
                    }
                    hierSteps.Add(new HierarchyStep() { 
                        ChildLevel = chLevelName, Name=stepName, ParentLevel=parLevelName, Predicate=predicName, HierarchyName=hierName,
                        IsForward = (direction.Equals(SparqlCommands.forwardPred)), IsOptional=optional.Equals("1"), IsPartial=partial.Equals("1")
                    });
                }
            }
            return hierSteps;
        }

        public List<HierarchyLevel> GetHierarchyLevels(string hierName, string dimName)
        {
            List<HierarchyLevel> hierLvls = new List<HierarchyLevel>();
            SparqlQuery mainQuery = null;
            SparqlQueryParser parser = new SparqlQueryParser();
            string cmdSparql = SparqlCommands.CmdPrefixes + string.Format(SparqlCommands.CmdHierarchyLevels, schemaGraphIRI, hierName, dimName);
            mainQuery = parser.ParseFromString(cmdSparql);
            mainQuery.Timeout = 7200 * 1000;

            SparqlRemoteEndpoint sre = new SparqlRemoteEndpoint(schemaSparqlEndpointUri);
            sre.Timeout = 7200 * 1000;
            RemoteQueryProcessor rqp = new RemoteQueryProcessor(sre);
            object obj = rqp.ProcessQuery(mainQuery);
            string str = mainQuery.ToString();
            Console.WriteLine(str);
            if (obj is SparqlResultSet)
            {
                SparqlResultSet rSet = (SparqlResultSet)obj;
                DataTable resultsTable = ConvertResultToDataTable(rSet);
                foreach (DataRow item in resultsTable.Rows)
                {
                    string lvlName = item["lvl"].ToString();
                    string lvlNum = item["lvlName"].ToString();
                    string lvlViewable = "1";
                    if (item["viewable"] != null && !string.IsNullOrEmpty(item["viewable"].ToString()))
                    {
                        lvlViewable = item["viewable"].ToString();
                    }                        
                    hierLvls.Add(new HierarchyLevel()
                    {
                        LevelComponent = lvlName, DimensionName = dimName, HierarchyName = hierName, LevelNum = lvlNum, IsViewable = lvlViewable.Equals("1")
                    });
                }
            }
            return hierLvls;
        }

        public List<MeasureInfo> GetMeasures()
        {
            //List<MeasureInfo> hierSteps = new List<MeasureInfo>();
            if (measures == null) 
                measures = new List<MeasureInfo>();

            //// not to run this query all the time
            if (measures.Count > 0)
                return measures;

            SparqlQuery mainQuery = null;
            SparqlQueryParser parser = new SparqlQueryParser();
            string cmdSparql = string.Format(SparqlCommands.CmdMeasuresInfo, schemaGraphIRI); ////SparqlCommands.CmdPrefixes + 
            mainQuery = parser.ParseFromString(cmdSparql);
            mainQuery.Timeout = 7200 * 1000;

            SparqlRemoteEndpoint sre = new SparqlRemoteEndpoint(schemaSparqlEndpointUri);
            sre.Timeout = 7200 * 1000;
            RemoteQueryProcessor rqp = new RemoteQueryProcessor(sre);
            object obj = rqp.ProcessQuery(mainQuery);
            string str = mainQuery.ToString();
            Console.WriteLine(str);
            if (obj is SparqlResultSet)
            {
                SparqlResultSet rSet = (SparqlResultSet)obj;
                DataTable resultsTable = ConvertResultToDataTable(rSet);
                foreach (DataRow item in resultsTable.Rows)
                {
                    
                    string msrName = item["msrName"].ToString();
                    string agrFnct = item["agrFnct"].ToString();
                    MeasureInfo mi = new MeasureInfo()
                    {
                        Name = msrName.Substring(msrName.LastIndexOfAny(new char[] { '#', '/' }) + 1),
                        AggFunction = agrFnct.Substring(agrFnct.LastIndexOfAny(new char[] { '#', '/' }) + 1).ToUpper()
                    };

                    if (item["direc"] != null && !string.IsNullOrEmpty(item["direc"].ToString()))
                    {
                        string direc = item["direc"].ToString();
                        mi.IsForward = (direc.Equals(SparqlCommands.forwardPred));
                    }
                    if (item["prdcName"] != null && !string.IsNullOrEmpty(item["prdcName"].ToString()))
                    {
                        mi.Predicate = item["prdcName"].ToString();
                    }

                    measures.Add(mi); //hierSteps.Add(mi);
                }
            }
            return measures;//hierSteps;
        }

        /// <summary>
        /// get flat lattice (without hierarchies)
        /// </summary>
        /// <param name="dimlist"></param>
        public List<List<string>> GetDimensionCombinations(List<string> dimList)
        {
            List<List<string>> flatLattice = new List<List<string>>();
            double count = Math.Pow(2, dimList.Count);
            for (int i = 1; i <= count - 1; i++)
            {
                List<string> newRound = new List<string>();
                string str = Convert.ToString(i, 2).PadLeft(dimList.Count, '0');
                for (int j = 0; j < str.Length; j++)
                {
                    if (str[j] == '1')
                    {
                        newRound.Add(dimList[j]);
                    }
                }
                flatLattice.Add(newRound);
            }
            return flatLattice;
        }

        /// <summary>
        /// 
        /// </summary>
        /// <param name="allDimLevels">this is a list containing the list of levels of every dimension</param>
        /// <returns></returns>
        public List<CuboidNode> GetPermutations(List<List<HierarchyLevel>> allDimLevels)
        {
            if (allDimLevels.Count == 1)//// just one dimension - every cuboidnode has just one CuboidNodePart
            {
                List<HierarchyLevel> parts = allDimLevels[0];
                List<CuboidNode> mergedList = parts.Select(c => new CuboidNode() { NodeParts = new List<HierarchyLevel>() { c } }).ToList();
                return mergedList;
            }
            else if (allDimLevels.Count == 2) //// two dimensions - every cuboidNode has two CuboidNodeParts
            {
                List<HierarchyLevel> firstList = allDimLevels[0];
                List<HierarchyLevel> secondList = allDimLevels[1];
                List<CuboidNode> mergedList = new List<CuboidNode>();
                foreach (HierarchyLevel i in firstList)
                {
                    foreach (HierarchyLevel j in secondList)
                    {
                        mergedList.Add(new CuboidNode() { NodeParts = new List<HierarchyLevel> { i, j } });
                    }
                }
                return mergedList;
            }
            else if (allDimLevels.Count > 2)//// three and more dimensions - recursively build it. second argument is recursively called function - contains CuboidNodes
            {
                //// first argument
                List<HierarchyLevel> firstArg = allDimLevels[0];
                List<List<HierarchyLevel>> reducedDimensions = allDimLevels.Select(c => c).ToList();
                reducedDimensions.Remove(firstArg);
                List<CuboidNode> secondArgument = GetPermutations(reducedDimensions);
                List<CuboidNode> mergedList = new List<CuboidNode>();
                foreach (HierarchyLevel i in firstArg)
                {
                    foreach (CuboidNode j in secondArgument)
                    {
                        List<HierarchyLevel> newParts = j.NodeParts.Select(c => c).ToList();
                        newParts.Add(i);
                        mergedList.Add(new CuboidNode() { NodeParts = newParts });
                    }
                }
                return mergedList;
            }
            return null;
        }

//        public string BuildSPARQLfromCuboid(CuboidNode cn, List<Hierarchy> allHiers)
//        {
//            StringBuilder sbWhere = new StringBuilder();
//            StringBuilder sbSelect = new StringBuilder();
//            StringBuilder sbGroupBy = new StringBuilder();
//            foreach (HierarchyLevel hl in cn.NodeParts)
//            {
//                string hierLvl = hl.LevelComponent;
//                Hierarchy h = allHiers.Where(c => c.DimensionName.Equals(hl.DimensionName)).FirstOrDefault();
//                List<HierarchyLevel> hierLevels = h.HierarchyLevels;

//                //// can be several steps. we need to find all possible paths and choose the best one!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

//                HierarchyStep step = h.HierarchySteps.Where(c => c.ParentLevel.Equals(hl.LevelNum)).FirstOrDefault();
//                string parentVar = hl.LevelComponent.Substring(hl.LevelComponent.LastIndexOfAny(new char[] { '#', '/' }) + 1);
//                sbSelect.Append("?" + parentVar + " ");
//                string childVar = string.Empty;
//                if (step.ChildLevel.Equals(SparqlCommands.observationPred))
//                    childVar = step.ChildLevel.Substring(step.ChildLevel.LastIndexOfAny(new char[] { '#', '/' }) + 1);
//                else
//                {
//                    HierarchyLevel childLvl = hierLevels.Where(c => c.LevelNum.Equals(step.ChildLevel)).FirstOrDefault();
//                    childVar = childLvl.LevelComponent.Substring(hl.LevelComponent.LastIndexOfAny(new char[] { '#', '/' }) + 1);
//                }
//                string predic = step.Predicate;
//                if (step.IsForward) sbWhere.AppendLine("?" + childVar + " <" + predic + "> " + "?" + parentVar + " .");
//                else sbWhere.AppendLine("?" + parentVar + " <" + predic + "> " + "?" + childVar + " .");
//                //// now add to SELECT sb
//                while (step != null && !step.ChildLevel.Equals(SparqlCommands.observationPred))
//                {
//                    step = h.HierarchySteps.Where(c => c.ParentLevel.Equals(step.ChildLevel)).FirstOrDefault();
//                    ////level
//                    HierarchyLevel parentLevel = h.HierarchyLevels.Where(c => c.LevelNum.Equals(step.ParentLevel)).FirstOrDefault();
//                    parentVar = parentLevel.LevelComponent.Substring(hl.LevelComponent.LastIndexOfAny(new char[] { '#', '/' }) + 1);
//                        //step.ParentLevel.Substring(step.ParentLevel.LastIndexOfAny(new char[] { '#', '/' }) + 1);
//                    if (step.ChildLevel.Equals(SparqlCommands.observationPred))
//                        childVar = step.ChildLevel.Substring(step.ChildLevel.LastIndexOfAny(new char[] { '#', '/' }) + 1);
//                    else
//                    {
//                        HierarchyLevel childLvl = hierLevels.Where(c => c.LevelNum.Equals(step.ChildLevel)).FirstOrDefault();
//                        childVar = childLvl.LevelComponent.Substring(hl.LevelComponent.LastIndexOfAny(new char[] { '#', '/' }) + 1);
//                    }
//                        //childVar = step.ChildLevel.Substring(step.ChildLevel.LastIndexOfAny(new char[] { '#', '/' }) + 1);
//                    predic = step.Predicate;
//                    if (step.IsForward) sbWhere.AppendLine("?" + childVar + " <" + predic + "> " + "?" + parentVar + " .");
//                    else sbWhere.AppendLine("?" + parentVar + " <" + predic + "> " + "?" + childVar + " .");
//                        //if (step.IsForward) sbWhere.AppendLine(childVar + " " + predic + " " + parentVar);
//                        //else sbWhere.AppendLine(parentVar + " " + predic + " " + childVar);
//                }
//            }

//            sbGroupBy.Append(sbSelect.ToString());

//            List<MeasureInfo> mesrs = GetMeasures();
//            foreach (MeasureInfo mes in mesrs)
//            {
//                if (!string.IsNullOrEmpty(mes.Predicate))
//                {
//                    if (mes.IsForward) sbWhere.AppendLine("?Observation" + " <" + mes.Predicate + "> " + "?" + mes.Name + " .");
//                    else sbWhere.AppendLine("?" + mes.Name + " <" + mes.Predicate + "> " + "?Observation .");

//                    sbSelect.Append("(" + mes.AggFunction + "(?" + mes.Name + ") AS ?" + mes.Name + mes.AggFunction + ") ");
//                }
//                else
//                    sbSelect.Append("(" + mes.AggFunction + "(?Observation) AS ?" + mes.Name + ") ");
//            }

//            string sparql = SparqlCommands.CmdPrefixes + string.Format(@"
//SELECT (COUNT(*) AS ?count)
//WHERE {{
//  SELECT {0}
//  WHERE {{
//    {1}
//  }}
//  GROUP BY {2}
//}}", sbSelect.ToString(), sbWhere.ToString(), sbGroupBy.ToString());
//            return sparql;
//        }

        public string BuildSPARQLfromCuboid_V3(CuboidNode cn, List<ViewPath> paths, List<Hierarchy> allHiers)
        {
            List<string> unionQueries = new List<string>();
            foreach (ViewPath vp in paths)
            {
                List<HierarchyStep> steps = vp.Dimensions.SelectMany(c => c.Steps).ToList();
                unionQueries.Add(BuildObservationSPARQLfromPathList(cn, steps, allHiers));
            }
            StringBuilder sbWhere = new StringBuilder();
            sbWhere.Append(" { ");
            sbWhere.Append(string.Join(" } UNION { ", unionQueries.ToArray()));
            sbWhere.Append(" } ");

            StringBuilder sbSelect = new StringBuilder();
            StringBuilder sbGroupBy = new StringBuilder();
            foreach (HierarchyLevel hl in cn.NodeParts)
            {
                string hierLvl = hl.LevelComponent;
                Hierarchy h = allHiers.Where(c => c.DimensionName.Equals(hl.DimensionName)).FirstOrDefault();
                List<HierarchyLevel> hierLevels = h.HierarchyLevels;

                HierarchyStep step = h.HierarchySteps.Where(c => c.ParentLevel.Equals(hl.LevelNum)).FirstOrDefault();
                string parentVar = hl.LevelComponent.Substring(hl.LevelComponent.LastIndexOfAny(new char[] { '#', '/' }) + 1);
                sbSelect.Append("?" + parentVar + " ");
                
            }

            sbGroupBy.Append(sbSelect.ToString());

            List<MeasureInfo> mesrs = GetMeasures();
            foreach (MeasureInfo mes in mesrs)
            {
                if (!string.IsNullOrEmpty(mes.Predicate))
                {
                    sbSelect.Append("(" + mes.AggFunction + "(?" + mes.Name + ") AS ?" + mes.Name + mes.AggFunction + ") ");
                }
                else
                    sbSelect.Append("(" + mes.AggFunction + "(?Observation) AS ?" + mes.Name + ") ");
            }

            string sparql = SparqlCommands.CmdPrefixes + string.Format(@"
SELECT (COUNT(*) AS ?count)
WHERE {{
  SELECT {0}
  WHERE {{
    {1}
  }}
  GROUP BY {2}
}}", sbSelect.ToString(), sbWhere.ToString(), sbGroupBy.ToString());
            return sparql;
        }

//        public List<string> BuildSPARQLfromCuboid_V2(CuboidNode cn, List<Hierarchy> allHiers)
//        {
//            List<List<HierarchyStep>> allPossPaths = new List<List<HierarchyStep>>();

//            foreach (HierarchyLevel hl in cn.NodeParts)
//            {
//                string hierLvl = hl.LevelComponent;
//                Hierarchy h = allHiers.Where(c => c.DimensionName.Equals(hl.DimensionName)).FirstOrDefault();
//                List<HierarchyLevel> hierLevels = h.HierarchyLevels;

//                //// can be several steps. we need to find all possible paths and choose the best one!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

//                List<HierarchyStep> steps = h.HierarchySteps.Where(c => c.ParentLevel.Equals(hl.LevelNum)).ToList();
//                foreach (HierarchyStep step in steps)
//                {

//                }

//                string parentVar = hl.LevelComponent.Substring(hl.LevelComponent.LastIndexOfAny(new char[] { '#', '/' }) + 1);
//                sbSelect.Append("?" + parentVar + " ");
//                string childVar = string.Empty;
//                if (step.ChildLevel.Equals(SparqlCommands.observationPred))
//                    childVar = step.ChildLevel.Substring(step.ChildLevel.LastIndexOfAny(new char[] { '#', '/' }) + 1);
//                else
//                {
//                    HierarchyLevel childLvl = hierLevels.Where(c => c.LevelNum.Equals(step.ChildLevel)).FirstOrDefault();
//                    childVar = childLvl.LevelComponent.Substring(hl.LevelComponent.LastIndexOfAny(new char[] { '#', '/' }) + 1);
//                }
//                string predic = step.Predicate;
//                if (step.IsForward) sbWhere.AppendLine("?" + childVar + " <" + predic + "> " + "?" + parentVar + " .");
//                else sbWhere.AppendLine("?" + parentVar + " <" + predic + "> " + "?" + childVar + " .");
//                //// now add to SELECT sb
//                while (step != null && !step.ChildLevel.Equals(SparqlCommands.observationPred))
//                {
//                    step = h.HierarchySteps.Where(c => c.ParentLevel.Equals(step.ChildLevel)).FirstOrDefault();
//                    ////level
//                    HierarchyLevel parentLevel = h.HierarchyLevels.Where(c => c.LevelNum.Equals(step.ParentLevel)).FirstOrDefault();
//                    parentVar = parentLevel.LevelComponent.Substring(hl.LevelComponent.LastIndexOfAny(new char[] { '#', '/' }) + 1);
//                    //step.ParentLevel.Substring(step.ParentLevel.LastIndexOfAny(new char[] { '#', '/' }) + 1);
//                    if (step.ChildLevel.Equals(SparqlCommands.observationPred))
//                        childVar = step.ChildLevel.Substring(step.ChildLevel.LastIndexOfAny(new char[] { '#', '/' }) + 1);
//                    else
//                    {
//                        HierarchyLevel childLvl = hierLevels.Where(c => c.LevelNum.Equals(step.ChildLevel)).FirstOrDefault();
//                        childVar = childLvl.LevelComponent.Substring(hl.LevelComponent.LastIndexOfAny(new char[] { '#', '/' }) + 1);
//                    }
//                    //childVar = step.ChildLevel.Substring(step.ChildLevel.LastIndexOfAny(new char[] { '#', '/' }) + 1);
//                    predic = step.Predicate;
//                    if (step.IsForward) sbWhere.AppendLine("?" + childVar + " <" + predic + "> " + "?" + parentVar + " .");
//                    else sbWhere.AppendLine("?" + parentVar + " <" + predic + "> " + "?" + childVar + " .");
//                    //if (step.IsForward) sbWhere.AppendLine(childVar + " " + predic + " " + parentVar);
//                    //else sbWhere.AppendLine(parentVar + " " + predic + " " + childVar);
//                }
//            }

//            StringBuilder sbWhere = new StringBuilder();
//            StringBuilder sbSelect = new StringBuilder();
            

//            string sparql = SparqlCommands.CmdPrefixes + string.Format(@"
//SELECT (COUNT(*) AS ?count)
//WHERE {{
//  SELECT {0} (COUNT(?Observation) as ?cnt)
//  WHERE {{
//    {1}
//  }}
//  GROUP BY {0}
//}}", sbSelect.ToString(), sbWhere.ToString());
//            return sparql;
//        }

        public List<ViewPath> GetPathPermutations(List<List<HierarchyPath>> allDimPaths)
        {
            if (allDimPaths.Count == 1)//// just one dimension - every cuboidnode has just one CuboidNodePart
            {
                List<HierarchyPath> parts = allDimPaths[0];
                List<ViewPath> mergedList = parts.Select(c => new ViewPath() { Dimensions = new List<HierarchyPath>() { c } }).ToList();
                return mergedList;
            }
            else if (allDimPaths.Count == 2) //// two dimensions - every cuboidNode has two CuboidNodeParts
            {
                List<HierarchyPath> firstList = allDimPaths[0];
                List<HierarchyPath> secondList = allDimPaths[1];
                List<ViewPath> mergedList = new List<ViewPath>();
                foreach (HierarchyPath i in firstList)
                {
                    foreach (HierarchyPath j in secondList)
                    {
                        mergedList.Add(new ViewPath() { Dimensions = new List<HierarchyPath> { i, j } });
                    }
                }
                return mergedList;
            }
            else if (allDimPaths.Count > 2)//// three and more dimensions - recursively build it. second argument is recursively called function - contains CuboidNodes
            {
                //// first argument
                List<HierarchyPath> firstArg = allDimPaths[0];
                List<List<HierarchyPath>> reducedDimensions = allDimPaths.Select(c => c).ToList();
                reducedDimensions.Remove(firstArg);
                List<ViewPath> secondArgument = GetPathPermutations(reducedDimensions);
                List<ViewPath> mergedList = new List<ViewPath>();
                foreach (HierarchyPath i in firstArg)
                {
                    foreach (ViewPath j in secondArgument)
                    {
                        List<HierarchyPath> newParts = j.Dimensions.Select(c => c).ToList();
                        newParts.Add(i);
                        mergedList.Add(new ViewPath() { Dimensions = newParts });
                    }
                }
                return mergedList;
            }
            return null;
        }

        public List<ViewPath> GetViewPaths(CuboidNode cn, List<Hierarchy> allHiers)
        {
            List<List<HierarchyPath>> allDimPaths = new List<List<HierarchyPath>>();
            foreach (HierarchyLevel hl in cn.NodeParts)
            {
                string hierLvl = hl.LevelComponent;
                Hierarchy h = allHiers.Where(c => c.DimensionName.Equals(hl.DimensionName)).FirstOrDefault();
                //List<HierarchyLevel> hierLevels = h.HierarchyLevels;
                List<HierarchyPath> theList = new List<HierarchyPath>();
                List<List<HierarchyStep>> allPaths = getPath_V2(h.HierarchySteps, hl.LevelNum);
                foreach (List<HierarchyStep> path in allPaths)
                {
                    theList.Add(new HierarchyPath() { Steps = path, HierarchyLevel = hl, HierarchyName = h.DimensionName });
                }

                //List<HierarchyStep> steps = h.HierarchySteps.Where(c => c.ParentLevel.Equals(hl.LevelNum)).ToList();
                //foreach (HierarchyStep step in steps)
                //{
                //    List<List<HierarchyStep>> allPaths = getPath_V2(h.HierarchySteps, step.ChildLevel);
                //    foreach (List<HierarchyStep> path in allPaths)
                //    {
                //        path.Add(step);
                //        theList.Add(new HierarchyPath() { Steps = path, HierarchyLevel = hl, HierarchyName = h.DimensionName });
                //    }
                //    //List<HierarchyStep> listElement = new List<HierarchyStep>() { step };
                //    //listElement.AddRange(getPath(h.HierarchySteps, step.ChildLevel));
                //    //theList.Add(new HierarchyPath() { Steps = listElement, HierarchyLevel = hl, HierarchyName = h.DimensionName });
                //}
                allDimPaths.Add(theList);
            }

            return GetPathPermutations(allDimPaths);
        }

        private List<HierarchyStep> getPath(List<HierarchyStep> steps, string levelNum)
        {
            List<HierarchyStep> stepsDown = steps.Where(c => c.ParentLevel.Equals(levelNum)).ToList();
            //// if the last level is reached
            if (stepsDown.Count == 1 && stepsDown[0].ChildLevel.Equals(SparqlCommands.observationPred))
                return stepsDown;
            else
            {
                foreach (HierarchyStep step in stepsDown)
                {
                    List<HierarchyStep> stepsOne = new List<HierarchyStep>();
                    stepsOne.Add(step);
                    stepsOne.AddRange(getPath(steps, step.ChildLevel));
                    return stepsOne;
                }
            }
            return stepsDown;
        }

        private List<List<HierarchyStep>> getPath_V2(List<HierarchyStep> steps, string levelNum)
        {
            List<HierarchyStep> stepsDown = steps.Where(c => c.ParentLevel.Equals(levelNum)).ToList();
            //// if the last level is reached
            if (stepsDown.Count == 1 && stepsDown[0].ChildLevel.Equals(SparqlCommands.observationPred))
                return new List<List<HierarchyStep>>(){ stepsDown };
            else
            {
                List<List<HierarchyStep>> allPaths = new List<List<HierarchyStep>>();
                foreach (HierarchyStep step in stepsDown)
                {
                    List<List<HierarchyStep>> previousSteps = getPath_V2(steps, step.ChildLevel);
                    if (previousSteps.Count == 0)
                        allPaths.Add(new List<HierarchyStep>() { step });
                    foreach (List<HierarchyStep> prevSteps in previousSteps)
                    {
                        prevSteps.Add(step);
                        allPaths.Add(prevSteps);
                    }
                }
                return allPaths;
            }
        }

        public string BuildCountSPARQLfromPathList(CuboidNode cn, List<HierarchyStep> steps, List<Hierarchy> allHiers)
        {
            StringBuilder sbWhere = new StringBuilder();
            StringBuilder sbSelect = new StringBuilder();
            foreach (HierarchyLevel hl in cn.NodeParts)
            {
                string hierLvl = hl.LevelComponent;
                Hierarchy h = allHiers.Where(c => c.DimensionName.Equals(hl.DimensionName)).FirstOrDefault();
                List<HierarchyLevel> hierLevels = h.HierarchyLevels;
                HierarchyStep step = steps.Where(c => c.ParentLevel.Equals(hl.LevelNum)).FirstOrDefault();
                string parentVar = hl.LevelComponent.Substring(hl.LevelComponent.LastIndexOfAny(new char[] { '#', '/' }) + 1);

                string childVar = string.Empty;
                if (step.ChildLevel.Equals(SparqlCommands.observationPred))
                    childVar = step.ChildLevel.Substring(step.ChildLevel.LastIndexOfAny(new char[] { '#', '/' }) + 1);
                else
                {
                    HierarchyLevel childLvl = hierLevels.Where(c => c.LevelNum.Equals(step.ChildLevel)).FirstOrDefault();
                    childVar = childLvl.LevelComponent.Substring(hl.LevelComponent.LastIndexOfAny(new char[] { '#', '/' }) + 1);
                }
                string predic = step.Predicate;
                if (!step.IsOptional) //// if the data is optional - there is no point in counting - not sure if the lack of data is because of scarcity or natural
                {
                    sbSelect.Append("?" + parentVar + " ");
                    if (step.IsForward) 
                        sbWhere.AppendLine("?" + childVar + " <" + predic + "> " + "?" + parentVar + " .");
                    else 
                        sbWhere.AppendLine("?" + parentVar + " <" + predic + "> " + "?" + childVar + " .");
                }
                else
                    Logging.Logger.log("Step is otpional: " + "Parent: " + parentVar + ", Predicate: " + predic + ", Child: " + childVar);
                while (step != null && !step.ChildLevel.Equals(SparqlCommands.observationPred))
                {
                    step = steps.Where(c => c.ParentLevel.Equals(step.ChildLevel)).FirstOrDefault();
                    ////level
                    HierarchyLevel parentLevel = h.HierarchyLevels.Where(c => c.LevelNum.Equals(step.ParentLevel)).FirstOrDefault();
                    parentVar = parentLevel.LevelComponent.Substring(hl.LevelComponent.LastIndexOfAny(new char[] { '#', '/' }) + 1);
                    //step.ParentLevel.Substring(step.ParentLevel.LastIndexOfAny(new char[] { '#', '/' }) + 1);
                    if (step.ChildLevel.Equals(SparqlCommands.observationPred))
                        childVar = step.ChildLevel.Substring(step.ChildLevel.LastIndexOfAny(new char[] { '#', '/' }) + 1);
                    else
                    {
                        HierarchyLevel childLvl = hierLevels.Where(c => c.LevelNum.Equals(step.ChildLevel)).FirstOrDefault();
                        childVar = childLvl.LevelComponent.Substring(hl.LevelComponent.LastIndexOfAny(new char[] { '#', '/' }) + 1);
                    }
                    //childVar = step.ChildLevel.Substring(step.ChildLevel.LastIndexOfAny(new char[] { '#', '/' }) + 1);
                    predic = step.Predicate;
                    if (step.IsForward) 
                        sbWhere.AppendLine("?" + childVar + " <" + predic + "> " + "?" + parentVar + " .");
                    else 
                        sbWhere.AppendLine("?" + parentVar + " <" + predic + "> " + "?" + childVar + " .");
                    //if (step.IsForward) sbWhere.AppendLine(childVar + " " + predic + " " + parentVar);
                    //else sbWhere.AppendLine(parentVar + " " + predic + " " + childVar);
                }
            }

            List<MeasureInfo> mesrs = GetMeasures();
            foreach (MeasureInfo mes in mesrs)
            {
                if (!string.IsNullOrEmpty(mes.Predicate))
                {
                    if (mes.IsForward) sbWhere.AppendLine("?Observation" + " <" + mes.Predicate + "> " + "?" + mes.Name + " .");
                    else sbWhere.AppendLine("?" + mes.Name + " <" + mes.Predicate + "> " + "?Observation .");

                    sbSelect.Append("?" + mes.Name + " ");
                }
                else
                    sbSelect.Append(" ?Observation ");
            }

            string sparql = SparqlCommands.CmdPrefixes + string.Format(@"
SELECT (COUNT(*) AS ?count)
WHERE {{
  SELECT {0}
  WHERE {{
    {1}
  }}
}}", sbSelect.ToString(), sbWhere.ToString());
            return sparql;
        }

        public string BuildObservationSPARQLfromPathList(CuboidNode cn, List<HierarchyStep> steps, List<Hierarchy> allHiers)
        {
            StringBuilder sbWhere = new StringBuilder();
            StringBuilder sbSelect = new StringBuilder();
            foreach (HierarchyLevel hl in cn.NodeParts)
            {
                string hierLvl = hl.LevelComponent;
                Hierarchy h = allHiers.Where(c => c.DimensionName.Equals(hl.DimensionName)).FirstOrDefault();
                List<HierarchyLevel> hierLevels = h.HierarchyLevels;
                HierarchyStep step = steps.Where(c => c.ParentLevel.Equals(hl.LevelNum)).FirstOrDefault();
                string parentVar = hl.LevelComponent.Substring(hl.LevelComponent.LastIndexOfAny(new char[] { '#', '/' }) + 1);

                string childVar = string.Empty;
                if (step.ChildLevel.Equals(SparqlCommands.observationPred))
                    childVar = step.ChildLevel.Substring(step.ChildLevel.LastIndexOfAny(new char[] { '#', '/' }) + 1);
                else
                {
                    HierarchyLevel childLvl = hierLevels.Where(c => c.LevelNum.Equals(step.ChildLevel)).FirstOrDefault();
                    childVar = childLvl.LevelComponent.Substring(hl.LevelComponent.LastIndexOfAny(new char[] { '#', '/' }) + 1);
                }
                string predic = step.Predicate;
                //if (!step.IsOptional) //// if the data is optional - there is no point in counting - not sure if the lack of data is because of scarcity or natural
                //{
                    sbSelect.Append("?" + parentVar + " ");
                    if (step.IsForward) sbWhere.AppendLine("?" + childVar + " <" + predic + "> " + "?" + parentVar + " .");
                    else sbWhere.AppendLine("?" + parentVar + " <" + predic + "> " + "?" + childVar + " .");
                //}
                //else
                //    Logging.Logger.log("Step is otpional: " + "Parent: " + parentVar + ", Predicate: " + predic + ", Child: " + childVar);
                while (step != null && !step.ChildLevel.Equals(SparqlCommands.observationPred))
                {
                    step = steps.Where(c => c.ParentLevel.Equals(step.ChildLevel)).FirstOrDefault();
                    ////level
                    HierarchyLevel parentLevel = h.HierarchyLevels.Where(c => c.LevelNum.Equals(step.ParentLevel)).FirstOrDefault();
                    parentVar = parentLevel.LevelComponent.Substring(hl.LevelComponent.LastIndexOfAny(new char[] { '#', '/' }) + 1);
                    if (step.ChildLevel.Equals(SparqlCommands.observationPred))
                        childVar = step.ChildLevel.Substring(step.ChildLevel.LastIndexOfAny(new char[] { '#', '/' }) + 1);
                    else
                    {
                        HierarchyLevel childLvl = hierLevels.Where(c => c.LevelNum.Equals(step.ChildLevel)).FirstOrDefault();
                        childVar = childLvl.LevelComponent.Substring(hl.LevelComponent.LastIndexOfAny(new char[] { '#', '/' }) + 1);
                    }
                    predic = step.Predicate;
                    if (step.IsForward) sbWhere.AppendLine("?" + childVar + " <" + predic + "> " + "?" + parentVar + " .");
                    else sbWhere.AppendLine("?" + parentVar + " <" + predic + "> " + "?" + childVar + " .");
                }
            }

            List<MeasureInfo> mesrs = GetMeasures();
            foreach (MeasureInfo mes in mesrs)
            {
                if (!string.IsNullOrEmpty(mes.Predicate))
                {
                    if (mes.IsForward) sbWhere.AppendLine("?Observation" + " <" + mes.Predicate + "> " + "?" + mes.Name + " .");
                    else sbWhere.AppendLine("?" + mes.Name + " <" + mes.Predicate + "> " + "?Observation .");

                    sbSelect.Append("?" + mes.Name + " ");
                }
                else
                    sbSelect.Append(" ?Observation ");
            }

            string sparql = string.Format(@"SELECT {0}
WHERE {{
  {1}
}}", sbSelect.ToString(), sbWhere.ToString(), dataGraphIRI);
            return sparql;
        }

        public string BuildCountSPARQLfromCuboid(CuboidNode cn, List<Hierarchy> allHiers)
        {
            StringBuilder sbWhere = new StringBuilder();
            StringBuilder sbSelect = new StringBuilder();
            foreach (HierarchyLevel hl in cn.NodeParts)
            {
                string hierLvl = hl.LevelComponent;
                Hierarchy h = allHiers.Where(c => c.DimensionName.Equals(hl.DimensionName)).FirstOrDefault();
                List<HierarchyLevel> hierLevels = h.HierarchyLevels;
                HierarchyStep step = h.HierarchySteps.Where(c => c.ParentLevel.Equals(hl.LevelNum)).FirstOrDefault();
                string parentVar = hl.LevelComponent.Substring(hl.LevelComponent.LastIndexOfAny(new char[] { '#', '/' }) + 1);
                
                string childVar = string.Empty;
                if (step.ChildLevel.Equals(SparqlCommands.observationPred))
                    childVar = step.ChildLevel.Substring(step.ChildLevel.LastIndexOfAny(new char[] { '#', '/' }) + 1);
                else
                {
                    HierarchyLevel childLvl = hierLevels.Where(c => c.LevelNum.Equals(step.ChildLevel)).FirstOrDefault();
                    childVar = childLvl.LevelComponent.Substring(hl.LevelComponent.LastIndexOfAny(new char[] { '#', '/' }) + 1);
                }
                string predic = step.Predicate;
                if (!step.IsOptional) //// if the data is optional - there is no point in counting - not sure if the lack of data is because of scarcity or natural
                {
                    sbSelect.Append("?" + parentVar + " ");
                    if (step.IsForward) sbWhere.AppendLine("?" + childVar + " <" + predic + "> " + "?" + parentVar + " .");
                    else sbWhere.AppendLine("?" + parentVar + " <" + predic + "> " + "?" + childVar + " .");
                }
                else
                    Logging.Logger.log("Step is otpional: " + "Parent: " + parentVar + ", Predicate: " + predic + ", Child: " + childVar);
                while (step != null && !step.ChildLevel.Equals(SparqlCommands.observationPred))
                {
                    step = h.HierarchySteps.Where(c => c.ParentLevel.Equals(step.ChildLevel)).FirstOrDefault();
                    ////level
                    HierarchyLevel parentLevel = h.HierarchyLevels.Where(c => c.LevelNum.Equals(step.ParentLevel)).FirstOrDefault();
                    parentVar = parentLevel.LevelComponent.Substring(hl.LevelComponent.LastIndexOfAny(new char[] { '#', '/' }) + 1);
                    //step.ParentLevel.Substring(step.ParentLevel.LastIndexOfAny(new char[] { '#', '/' }) + 1);
                    if (step.ChildLevel.Equals(SparqlCommands.observationPred))
                        childVar = step.ChildLevel.Substring(step.ChildLevel.LastIndexOfAny(new char[] { '#', '/' }) + 1);
                    else
                    {
                        HierarchyLevel childLvl = hierLevels.Where(c => c.LevelNum.Equals(step.ChildLevel)).FirstOrDefault();
                        childVar = childLvl.LevelComponent.Substring(hl.LevelComponent.LastIndexOfAny(new char[] { '#', '/' }) + 1);
                    }
                    //childVar = step.ChildLevel.Substring(step.ChildLevel.LastIndexOfAny(new char[] { '#', '/' }) + 1);
                    predic = step.Predicate;
                    if (step.IsForward) sbWhere.AppendLine("?" + childVar + " <" + predic + "> " + "?" + parentVar + " .");
                    else sbWhere.AppendLine("?" + parentVar + " <" + predic + "> " + "?" + childVar + " .");
                    //if (step.IsForward) sbWhere.AppendLine(childVar + " " + predic + " " + parentVar);
                    //else sbWhere.AppendLine(parentVar + " " + predic + " " + childVar);
                }
            }

            List<MeasureInfo> mesrs = GetMeasures();
            foreach (MeasureInfo mes in mesrs)
            {
                if (!string.IsNullOrEmpty(mes.Predicate))
                {
                    if (mes.IsForward) sbWhere.AppendLine("?Observation" + " <" + mes.Predicate + "> " + "?" + mes.Name + " .");
                    else sbWhere.AppendLine("?" + mes.Name + " <" + mes.Predicate + "> " + "?Observation .");

                    sbSelect.Append("?" + mes.Name + " ");
                }
                else
                    sbSelect.Append(" ?Observation ");
            }

            string sparql = SparqlCommands.CmdPrefixes + string.Format(@"
SELECT (COUNT(*) AS ?count)
WHERE {{
  SELECT {0}
  WHERE {{
    {1}
  }}
}}", sbSelect.ToString(), sbWhere.ToString(), dataGraphIRI);
            return sparql;
        }

        /// <summary>
        /// i need to get it from cuboid. cannot estimate it every time from scratch - need to save the size of the view
        /// </summary>
        /// <param name="cn"></param>
        /// <param name="allHiers"></param>
        /// <returns></returns>
        public string BuildCountSPARQLfromCuboidPreviousLevel(CuboidNode cn, List<Hierarchy> allHiers)
        {
            StringBuilder sbWhere = new StringBuilder();
            StringBuilder sbSelect = new StringBuilder();
            foreach (HierarchyLevel hl in cn.NodeParts)
            {
                string hierLvl = hl.LevelComponent;
                Hierarchy h = allHiers.Where(c => c.DimensionName.Equals(hl.DimensionName)).FirstOrDefault();
                List<HierarchyLevel> hierLevels = h.HierarchyLevels;
                HierarchyStep step = h.HierarchySteps.Where(c => c.ParentLevel.Equals(hl.LevelNum)).FirstOrDefault();
                string parentVar = hl.LevelComponent.Substring(hl.LevelComponent.LastIndexOfAny(new char[] { '#', '/' }) + 1);
                //if (step.IsOptional) //// only if mandatory
                //    sbSelect.Append("?" + parentVar + " ");
                string childVar = string.Empty;
                string predic = step.Predicate;
                if (step.ChildLevel.Equals(SparqlCommands.observationPred))
                {
                    //return GetObservationCountQuery();

                    childVar = step.ChildLevel.Substring(step.ChildLevel.LastIndexOfAny(new char[] { '#', '/' }) + 1);
                    if (step.IsForward) sbWhere.AppendLine("?" + childVar + " <" + predic + "> " + "?" + parentVar + " .");
                    else sbWhere.AppendLine("?" + parentVar + " <" + predic + "> " + "?" + childVar + " .");
                }
                //else
                //{
                //    HierarchyLevel childLvl = hierLevels.Where(c => c.LevelNum.Equals(step.ChildLevel)).FirstOrDefault();
                //    childVar = childLvl.LevelComponent.Substring(hl.LevelComponent.LastIndexOfAny(new char[] { '#', '/' }) + 1);
                //}
                
                //if (step.IsOptional) //// only if mandatory
                //{
                //    if (step.IsForward) sbWhere.AppendLine("?" + childVar + " <" + predic + "> " + "?" + parentVar + " .");
                //    else sbWhere.AppendLine("?" + parentVar + " <" + predic + "> " + "?" + childVar + " .");
                //}
                //// now add to SELECT sb
                while (step != null && !step.ChildLevel.Equals(SparqlCommands.observationPred))
                {
                    step = h.HierarchySteps.Where(c => c.ParentLevel.Equals(step.ChildLevel)).FirstOrDefault();
                    ////level
                    HierarchyLevel parentLevel = h.HierarchyLevels.Where(c => c.LevelNum.Equals(step.ParentLevel)).FirstOrDefault();
                    parentVar = parentLevel.LevelComponent.Substring(hl.LevelComponent.LastIndexOfAny(new char[] { '#', '/' }) + 1);
                    //step.ParentLevel.Substring(step.ParentLevel.LastIndexOfAny(new char[] { '#', '/' }) + 1);
                    if (step.ChildLevel.Equals(SparqlCommands.observationPred))
                        childVar = step.ChildLevel.Substring(step.ChildLevel.LastIndexOfAny(new char[] { '#', '/' }) + 1);
                    else
                    {
                        HierarchyLevel childLvl = hierLevels.Where(c => c.LevelNum.Equals(step.ChildLevel)).FirstOrDefault();
                        childVar = childLvl.LevelComponent.Substring(hl.LevelComponent.LastIndexOfAny(new char[] { '#', '/' }) + 1);
                    }
                    //childVar = step.ChildLevel.Substring(step.ChildLevel.LastIndexOfAny(new char[] { '#', '/' }) + 1);
                    predic = step.Predicate;
                    if (step.IsForward) sbWhere.AppendLine("?" + childVar + " <" + predic + "> " + "?" + parentVar + " .");
                    else sbWhere.AppendLine("?" + parentVar + " <" + predic + "> " + "?" + childVar + " .");
                    //if (step.IsForward) sbWhere.AppendLine(childVar + " " + predic + " " + parentVar);
                    //else sbWhere.AppendLine(parentVar + " " + predic + " " + childVar);
                }
            }

            List<MeasureInfo> mesrs = GetMeasures();
            foreach (MeasureInfo mes in mesrs)
            {
                if (!string.IsNullOrEmpty(mes.Predicate))
                {
                    if (mes.IsForward) sbWhere.AppendLine("?Observation" + " <" + mes.Predicate + "> " + "?" + mes.Name + " .");
                    else sbWhere.AppendLine("?" + mes.Name + " <" + mes.Predicate + "> " + "?Observation .");

                    sbSelect.Append("?" + mes.Name + " ");
                }
                else
                    sbSelect.Append(" ?Observation ");
            }

            string sparql = SparqlCommands.CmdPrefixes + string.Format(@"
SELECT (COUNT(*) AS ?count)
WHERE {{
  SELECT {0}
  WHERE {{
    {1}
  }}
}}", sbSelect.ToString(), sbWhere.ToString());
//            string sparql = SparqlCommands.CmdPrefixes + string.Format(@"
//SELECT (COUNT(*) AS ?count)
//WHERE {{
//  SELECT {0}
//  FROM <{2}>
//  WHERE {{
//    GRAPH <{2}> {{
//      {1}
//    }}
//  }}
//}}", sbSelect.ToString(), sbWhere.ToString(), dataGraphIRI);
            return sparql;
        }

        ///// <summary>
        ///// This one is for the case when completeness is specified in the schema 
        ///// </summary>
        ///// <param name="cn"></param>
        ///// <param name="allHiers"></param>
        ///// <returns></returns>
        //public bool IsCuboidNodeComplete(CuboidNode cn, List<Hierarchy> allHiers)
        //{
        //    foreach (HierarchyLevel hl in cn.NodeParts)
        //    {
        //        string hierLvl = hl.LevelComponent;
        //        Hierarchy h = allHiers.Where(c => c.DimensionName.Equals(hl.DimensionName)).FirstOrDefault();
        //        List<HierarchyLevel> hierLevels = h.HierarchyLevels;
        //        HierarchyStep step = h.HierarchySteps.Where(c => c.ParentLevel.Equals(hl.LevelNum)).FirstOrDefault();
        //        if (!step.IsOptional) //// only if mandatory
        //        {
        //            return false;
        //        }
        //    }
        //    return true;
        //}

        ///// <summary>
        ///// This one is for the case when completeness is calculated by SPARQL queries
        ///// </summary>
        ///// <param name="cn"></param>
        ///// <param name="allHiers"></param>
        ///// <returns></returns>
        //public bool IsCuboidNodeComplete_V2(CuboidNode cn, List<Hierarchy> allHiers)
        //{
        //    int i = 0;
        //    string sparql = BuildCountSPARQLfromCuboid(cn, allHiers);
        //    Logging.Logger.logSeparately("Cuboid node COUNT query " + i.ToString() + ": " + sparql);
        //    int sizeWith = GetViewSize(sparql);
        //    sparql = BuildCountSPARQLfromCuboidPreviousLevel(cn, allHiers);
        //    Logging.Logger.logSeparately("Cuboid node COUNT EXCEPT query " + i.ToString() + ": " + sparql);
        //    int sizeWithout = GetViewSize(sparql);
        //    double completeness = (((double)sizeWith) / sizeWithout);
        //    Logging.Logger.logSeparately("Cuboid node query " + i.ToString() + " is " + ((completeness == 1.0) ? "complete" : "incomplete"));
        //    return (completeness == 1.0);

        //    //foreach (HierarchyLevel hl in cn.NodeParts)
        //    //{
        //    //    string hierLvl = hl.LevelComponent;
        //    //    Hierarchy h = allHiers.Where(c => c.DimensionName.Equals(hl.DimensionName)).FirstOrDefault();
        //    //    List<HierarchyLevel> hierLevels = h.HierarchyLevels;
        //    //    HierarchyStep step = h.HierarchySteps.Where(c => c.ParentLevel.Equals(hl.LevelNum)).FirstOrDefault();
        //    //    //if (!step.IsMandatory) //// only if mandatory
        //    //    //{
        //    //    //    return false;
        //    //    //}

        //    //}
        //    //return true;
        //}

        /// <summary>
        /// This one is for the case when completeness is calculated by SPARQL queries
        /// </summary>
        /// <param name="cn"></param>
        /// <param name="allHiers"></param>
        /// <returns></returns>
        public void SetViewCompleteness(View v, List<Hierarchy> allHiers)
        {
            CuboidNode cn = v.Node;
            int i = 0;
            string sparql = BuildCountSPARQLfromCuboid(cn, allHiers);
            Logging.Logger.logSeparately("Cuboid node COUNT query " + i.ToString() + ": " + sparql);
            int sizeWith = GetViewSize(sparql);
            sparql = BuildCountSPARQLfromCuboidPreviousLevel(cn, allHiers);
            Logging.Logger.logSeparately("Cuboid node COUNT EXCEPT query " + i.ToString() + ": " + sparql);
            int sizeWithout = GetViewSize(sparql);
            double completeness = (((double)sizeWith) / sizeWithout);
            Logging.Logger.logSeparately("Cuboid node query " + i.ToString() + " is " + ((completeness == 1.0) ? "complete" : "incomplete"));
            v.Completeness = completeness;
        }

        public int GetObservationCount()
        {
            string sparqlCmdStr = GetObservationCountQuery();
            return GetViewSize(sparqlCmdStr);
        }

        /// <summary>
        /// check if matView is helpful in answering testView
        /// </summary>
        /// <param name="matView"></param>
        /// <param name="testView"></param>
        /// <returns></returns>
        public bool IsViewHelpful(View matView, View testView, List<Hierarchy> allHiers)
        {
            if (testView.Node.NodeParts.Count > matView.Node.NodeParts.Count) //// more parts in the lattice node - not possible to use
                return false;

            foreach (HierarchyLevel nodePart in testView.Node.NodeParts)
            {
                if (!matView.Node.NodeParts.Select(c => c.DimensionName).Contains(nodePart.DimensionName))//// no such lattice node part in matview - not possible to use
                    return false;
                //// get the cuboidPart from matview and compare with testview. if level is less - not possible to use (deeper in the hierarchy)
                HierarchyLevel matViewPart = matView.Node.NodeParts.Where(c => c.DimensionName.Equals(nodePart.DimensionName)).FirstOrDefault();

                //// is it possible to get it from previous steps?
                Hierarchy h = allHiers.Where(c => c.DimensionName.Equals(nodePart.DimensionName)).FirstOrDefault();
                List<HierarchyStep> pathSteps = GetPathSteps(h.HierarchySteps, matViewPart.LevelNum, nodePart.LevelNum); 
                //if (matViewPart.Level.Level < nodePart.Level.Level)
                if(pathSteps == null || pathSteps.Count == 0 || (pathSteps.Where(c => c.IsOptional).FirstOrDefault() != null))
                    return false;
            }
            return true;
        }

        /// <summary>
        /// check if matView is helpful in answering testView
        /// </summary>
        /// <param name="matView"></param>
        /// <param name="testView"></param>
        /// <returns></returns>
        public bool IsViewHelpful_V2(View matView, View testView, List<Hierarchy> allHiers, double completeness)
        {
            if (testView.Node.NodeParts.Count > matView.Node.NodeParts.Count) //// more parts in the lattice node - not possible to use
                return false;
            
            foreach (HierarchyLevel nodePart in testView.Node.NodeParts)
            {
                if (!matView.Node.NodeParts.Select(c => c.DimensionName).Contains(nodePart.DimensionName))//// no such lattice node part in matview - not possible to use
                    return false;
            }
            
            if (testView.SelectedPath != null && matView.SelectedPath != null)
            {
                List<HierarchyStep> mvPaths = matView.SelectedPath.Dimensions.SelectMany(c => c.Steps).ToList();
                List<HierarchyStep> testPaths = testView.SelectedPath.Dimensions.SelectMany(c => c.Steps).ToList();
                foreach (HierarchyStep stp in mvPaths)
                {
                    if (!testPaths.Contains(stp))//// matview contains steps that are not present in test view
                        return false;
                }
            }
            else if (matView.SelectedPath != null)
            {
                bool isPoss = true;
                List<HierarchyStep> mvSteps = matView.SelectedPath.Dimensions.SelectMany(c => c.Steps).ToList();
                foreach (ViewPath testPath in testView.Paths)
                {
                    List<HierarchyStep> testSteps = testPath.Dimensions.SelectMany(c => c.Steps).ToList();
                    foreach (HierarchyStep stp in mvSteps)
                    {
                        if (!testSteps.Contains(stp))//// matview contains steps that are not present in test view
                            isPoss = false;
                    }
                    if (isPoss && (matView.SelectedPath.Completeness * testPath.Completeness > completeness))
                        return true;
                }
            }

            //// check possible paths of the test view and the completeness level.
            foreach (HierarchyLevel nodePart in testView.Node.NodeParts)
            {
                //// get the cuboidPart from matview and compare with testview. if level is less - not possible to use (deeper in the hierarchy)
                HierarchyLevel matViewPart = matView.Node.NodeParts.Where(c => c.DimensionName.Equals(nodePart.DimensionName)).FirstOrDefault();

                //// is it possible to get it from previous steps?
                Hierarchy h = allHiers.Where(c => c.DimensionName.Equals(nodePart.DimensionName)).FirstOrDefault();
                List<HierarchyStep> pathSteps = GetPathSteps(h.HierarchySteps, matViewPart.LevelNum, nodePart.LevelNum);
                //if (matViewPart.Level.Level < nodePart.Level.Level)
                if (pathSteps == null || pathSteps.Count == 0)
                    return false;
            }

            foreach (ViewPath mvPath in matView.Paths)
            {
                foreach (ViewPath testPath in testView.Paths)
                {
                    bool isPoss = true;
                    List<HierarchyStep> mvSteps = mvPath.Dimensions.SelectMany(c => c.Steps).ToList();
                    List<HierarchyStep> testSteps = testPath.Dimensions.SelectMany(c => c.Steps).ToList();
                    foreach (HierarchyStep stp in mvSteps)
                    {
                        if (!testSteps.Contains(stp))//// matview contains steps that are not present in test view
                            isPoss = false;
                    }
                    if (isPoss && (mvPath.Completeness * testPath.Completeness > completeness))
                        return true;
                }
            }
            
            return false;
        }

        public bool IsViewHelpful_V3(View matView, View testView, List<Hierarchy> allHiers)
        {
            if (testView.Node.NodeParts.Count > matView.Node.NodeParts.Count) //// more parts in the lattice node - not possible to use
                return false;

            foreach (HierarchyLevel nodePart in testView.Node.NodeParts)
            {
                if (!matView.Node.NodeParts.Select(c => c.DimensionName).Contains(nodePart.DimensionName))//// no such lattice node part in matview - not possible to use
                    return false;
            }

            List<ViewPath> mvViewPaths = matView.Paths;
            List<ViewPath> testViewPaths = testView.Paths;
            

            //// check possible paths of the test view and the completeness level.
            foreach (HierarchyLevel nodePart in testView.Node.NodeParts)
            {
                //// get the cuboidPart from matview and compare with testview. if level is less - not possible to use (deeper in the hierarchy)
                HierarchyLevel matViewPart = matView.Node.NodeParts.Where(c => c.DimensionName.Equals(nodePart.DimensionName)).FirstOrDefault();

                //// is it possible to get it from previous steps?
                Hierarchy h = allHiers.Where(c => c.DimensionName.Equals(nodePart.DimensionName)).FirstOrDefault();
                List<HierarchyStep> pathSteps = GetPathSteps(h.HierarchySteps, matViewPart.LevelNum, nodePart.LevelNum);
                //if (matViewPart.Level.Level < nodePart.Level.Level)
                if (matViewPart.LevelNum.Equals(nodePart.LevelNum))  //// the same level
                    continue;
                if (pathSteps == null )//|| pathSteps.Count == 0
                    return false;
            }

            bool isPoss = true;
            foreach (ViewPath testPath in testView.Paths)
            {
                List<string> dimNames = testPath.Dimensions.Select(c => c.HierarchyName).ToList();
                isPoss = false;
                foreach (ViewPath mvPath in matView.Paths)
                {
                    //// dimension names!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!\

                    List<HierarchyPath> mvPaths = mvPath.Dimensions.Where(c => dimNames.Contains(c.HierarchyName)).ToList();
                    List<HierarchyStep> mvSteps = mvPaths.SelectMany(c => c.Steps).ToList(); //mvPath.Dimensions.SelectMany(c => c.Steps).ToList();
                    List<HierarchyStep> testSteps = testPath.Dimensions.SelectMany(c => c.Steps).ToList();
                    if (mvSteps.Except(testSteps).Count() == 0)
                        isPoss = true;
                    else if(mvSteps.Intersect(testSteps).Count() == mvSteps.Count)
                        isPoss = true;
                }
                if (!isPoss)
                    break;
            }

            return isPoss;
        }

        public int GetViewSize(string sparqlViewCountQuery)
        {
            List<string> dimensionNames = new List<string>();
            SparqlQuery mainQuery = null;
            SparqlQueryParser parser = new SparqlQueryParser();
            string cmdSparql = sparqlViewCountQuery;//SparqlCommands.CmdPrefixes + string.Format(SparqlCommands.CmdDimensionHierarchy, dimName);
            mainQuery = parser.ParseFromString(cmdSparql);
            mainQuery.Timeout = 7200 * 1000;

            SparqlRemoteEndpoint sre = new SparqlRemoteEndpoint(sparqlEndpointUri);
            sre.Timeout = 7200 * 1000;
            RemoteQueryProcessor rqp = new RemoteQueryProcessor(sre);
            string str = mainQuery.ToString();
            Console.WriteLine(str);
            object obj = rqp.ProcessQuery(mainQuery);
           
            if (obj is SparqlResultSet)
            {
                SparqlResultSet rSet = (SparqlResultSet)obj;
                if (rSet != null && rSet.Count > 0)
                {
                    SparqlResult result = rSet[0];
                    string item = result.Variables.ToList()[0];
                    INode nd;
                    string data = string.Empty;
                    if (result.TryGetValue(item, out nd))
                    {
                        if (nd != null)
                        {
                            ILiteralNode node = ((ILiteralNode)nd);
                            if (!((ILiteralNode)nd).Value.Equals("NAN"))
                                data = ((ILiteralNode)nd).Value;
                            else
                                data = "0";
                        }
                    }
                    int res = 0;
                    if (Int32.TryParse(data, out res))
                        return res;
                }
            }
            return 0;
        }

        public int GetAllTriplesCount()
        {
            string sparqlCmdStr = "SELECT (COUNT(*) AS ?count) FROM <{0}> WHERE {{ ?s ?p ?o . }}";
            //string sparqlCmdStr = "SELECT (COUNT(*) AS ?count) FROM <{0}> WHERE {{ ?s a <http://ssb.org/schemas/Lineorder> . }}";
            return GetViewSize(string.Format(sparqlCmdStr, dataGraphIRI));
        }

        /// <summary>
        /// get the benefit of a view for a list of all views
        /// materializedViews are the views which are already selected
        /// </summary>
        /// <param name="v"></param>
        /// <param name="allViews"></param>
        /// <returns></returns>
        public int GetViewBenefit(View testView, List<View> allViews, List<Hierarchy> allHiers)
        {
            //// benefit is at least the benefit of one materialized view (maybe not null at the beginning, need to think what it is equal to)
            int benefit = 0;
            ////for each lattice node check if view is useful and get its benefit.
            foreach (View vItem in allViews)
            {
                if (IsViewHelpful(testView, vItem, allHiers) && vItem.Cost > testView.Size) ////if testView can be used for answering vItem, and testView is cheaper than vItems method 
                {
                    benefit += vItem.Cost - testView.Size; //// how much is saved when vItem will be counted from testView
                }
            }
            return benefit;
        }

        /// <summary>
        /// get the benefit of a view for a list of all views
        /// materializedViews are the views which are already selected
        /// </summary>
        /// <param name="v"></param>
        /// <param name="allViews"></param>
        /// <returns></returns>
        public int GetViewBenefit_V2(View testView, List<View> allViews, List<Hierarchy> allHiers, double completeness)
        {
            //// benefit is at least the benefit of one materialized view (maybe not null at the beginning, need to think what it is equal to)
            int benefit = 0;
            ////for each lattice node check if view is useful and get its benefit.
            foreach (View vItem in allViews)
            {
                if (IsViewHelpful_V2(testView, vItem, allHiers, completeness) && vItem.Cost > testView.Size) ////if testView can be used for answering vItem, and testView is cheaper than vItems method 
                {
                    benefit += vItem.Cost - testView.Size; //// how much is saved when vItem will be counted from testView
                }
            }
            return benefit;
        }

        /// <summary>
        /// get the benefit of a view for a list of all views
        /// materializedViews are the views which are already selected
        /// </summary>
        /// <param name="v"></param>
        /// <param name="allViews"></param>
        /// <returns></returns>
        public int GetViewBenefit_V3(View testView, List<View> allViews, List<Hierarchy> allHiers)
        {
            //// benefit is at least the benefit of one materialized view (maybe not null at the beginning, need to think what it is equal to)
            int benefit = 0;
            ////for each lattice node check if view is useful and get its benefit.
            foreach (View vItem in allViews)
            {
                if (IsViewHelpful_V3(testView, vItem, allHiers) && vItem.Cost > testView.Size) ////if testView can be used for answering vItem, and testView is cheaper than vItems method 
                {
                    benefit += vItem.Cost - testView.Size; //// how much is saved when vItem will be counted from testView
                }
                //// self benefit
                if(testView.Equals(vItem)) benefit += vItem.Cost - testView.Size;
            }
            return benefit;
        }

        /// <summary>
        /// get the benefit of a view for a list of all views
        /// materializedViews are the views which are already selected
        /// </summary>
        /// <param name="v"></param>
        /// <param name="allViews"></param>
        /// <returns></returns>
        public double GetViewBenefit_V4(View testView, List<View> allViews, List<Hierarchy> allHiers)
        {
            //// benefit is at least the benefit of one materialized view (maybe not null at the beginning, need to think what it is equal to)
            int numOfViews = 1;
            double benefit = 0;
            ////for each lattice node check if view is useful and get its benefit.
            foreach (View vItem in allViews)
            {
                if (IsViewHelpful_V3(testView, vItem, allHiers) && vItem.Cost > testView.Size) ////if testView can be used for answering vItem, and testView is cheaper than vItems method 
                {
                    benefit += (double)(vItem.Cost - testView.Size); //// how much is saved when vItem will be counted from testView
                    numOfViews++;
                }
                //// self benefit
                if (testView.Equals(vItem)) benefit += (double)(vItem.Cost - testView.Size) ;//vItem.Cost - testView.Size;
            }
            return benefit/testView.Cost; //((numOfViews != 0) ? benefit*numOfViews : benefit); //(benefit/);
        }

        /// <summary>
        /// returns the vew from already materialized views that is able to answer this query. null if such a view is not found
        /// </summary>
        /// <param name="testView"></param>
        /// <param name="materializedViews"></param>
        /// <returns></returns>
        public View GetBestMatViewForView(View testView, List<View> materializedViews, List<Hierarchy> allHiers)
        {
            if (materializedViews == null || materializedViews.Count == 0)
                return null;
            View retView = null;
            int minViewTuples = 0;
            foreach (View matView in materializedViews)
            {
                if (IsViewHelpful(matView, testView, allHiers)) //test view can be answered by already materialized view
                {
                    if (minViewTuples == 0)
                    {
                        retView = matView;
                        minViewTuples = matView.Size;
                    }
                    else if (matView.Size < minViewTuples)
                    {
                        retView = matView;
                        minViewTuples = matView.Size;
                    }
                }
            }
            return retView;
        }

        /// <summary>
        /// returns the vew from already materialized views that is able to answer this query. null if such a view is not found
        /// </summary>
        /// <param name="testView"></param>
        /// <param name="materializedViews"></param>
        /// <returns></returns>
        public View GetBestMatViewForView_V2(View testView, List<View> materializedViews, List<Hierarchy> allHiers, double completeness)
        {
            if (materializedViews == null || materializedViews.Count == 0)
                return null;
            View retView = null;
            int minViewTuples = 0;
            foreach (View matView in materializedViews)
            {
                if (IsViewHelpful_V2(matView, testView, allHiers, completeness)) //test view can be answered by already materialized view
                {
                    if (minViewTuples == 0)
                    {
                        retView = matView;
                        minViewTuples = matView.Size;
                    }
                    else if (matView.Size < minViewTuples)
                    {
                        retView = matView;
                        minViewTuples = matView.Size;
                    }
                }
            }
            return retView;
        }

        /// <summary>
        /// returns the vew from already materialized views that is able to answer this query. null if such a view is not found
        /// </summary>
        /// <param name="testView"></param>
        /// <param name="materializedViews"></param>
        /// <returns></returns>
        public View GetBestMatViewForView_V3(View testView, List<View> materializedViews, List<Hierarchy> allHiers)
        {
            if (materializedViews == null || materializedViews.Count == 0)
                return null;
            View retView = null;
            int minViewTuples = 0;
            foreach (View matView in materializedViews)
            {
                if (IsViewHelpful_V3(matView, testView, allHiers)) //test view can be answered by already materialized view
                {
                    if (minViewTuples == 0)
                    {
                        retView = matView;
                        minViewTuples = matView.Size;
                    }
                    else if (matView.Size < minViewTuples)
                    {
                        retView = matView;
                        minViewTuples = matView.Size;
                    }
                }
                if (testView.Equals(matView)) 
                    return matView;
            }
            return retView;
        }

        public string GetViewBenefitForLogging(View testView, List<View> allViews, List<Hierarchy> allHiers)
        {
            StringBuilder sb = new StringBuilder();
            sb.AppendLine("Calculating view benefit for view " + testView.ToString());
            int benefit = 0;
            ////for each lattice node check if view is useful and get its benefit.
            foreach (View vItem in allViews)
            {
                if (IsViewHelpful(testView, vItem, allHiers) && vItem.Cost > testView.Size) ////if testView can be used for answering vItem, and testView is cheaper than vItems method 
                {
                    benefit += vItem.Cost - testView.Size; //// how much is saved when vItem will be counted from testView
                    sb.AppendLine(string.Format("for view {0}, with cost {1}, the benefit of matView is {2}", vItem.ToString(), vItem.Cost, (vItem.Cost - testView.Size)));
                }
            }
            return sb.ToString();
        }

        /// <summary>
        /// checks all preceding views to see if there is a view directly preceding the current view v
        /// </summary>
        /// <param name="v"></param>
        /// <param name="allKnownViews"></param>
        /// <returns></returns>
        public List<View> GetDirectlyPrecedingViews(View v, List<View> allKnownViews, List<Hierarchy> allHiers)
        {
            List<View> retViews = new List<View>();
            List<ViewPath> vvp = v.Paths;
            foreach (View view in allKnownViews)
            {
                foreach (ViewPath knownViewPath in view.Paths)
                {
                    foreach (ViewPath testViewPath in vvp)
                    {
                        if(knownViewPath.Dimensions.Count == testViewPath.Dimensions.Count)
                        {
                            //// if the difference between path is just one element
                            if (knownViewPath.Dimensions.SelectMany(c => c.Steps).ToList().Except(testViewPath.Dimensions.SelectMany(c => c.Steps).ToList()).Count() == 0 &&
                                testViewPath.Dimensions.SelectMany(c => c.Steps).ToList().Except(knownViewPath.Dimensions.SelectMany(c => c.Steps).ToList()).Count() == 1)
                                if (!retViews.Contains(view))
                                    retViews.Add(view);
                        }
                    }
                }
            }
            return retViews;
        }

        #region helper methods
        private static DataTable ConvertResultToDataTable(SparqlResultSet rset)
        {
            DataTable tbl = new DataTable();
            int rowNr = 0;
            foreach (SparqlResult result in rset)
            {
                List<string> vals = new List<string>();
                foreach (string item in result.Variables)
                {
                    if (rowNr == 0)
                    {
                        INode nd;
                        if (result.TryGetValue(item, out nd))
                        {
                            if (nd != null)
                            {
                                switch (nd.NodeType)
                                {
                                    case NodeType.Uri:
                                        tbl.Columns.Add(item);
                                        break;
                                    case NodeType.Literal:
                                        //You may want to inspect the DataType and Language properties and generate
                                        //a different string here
                                        ILiteralNode node = ((ILiteralNode)nd);
                                        //if (node.DataType != null && (node.DataType.Equals(UriFactory.Create("http://www.w3.org/2001/XMLSchema#decimal"))))
                                        //    tbl.Columns.Add(new DataColumn(item, typeof(float)));
                                        //else if (node.DataType != null && (node.DataType.Equals(UriFactory.Create("http://www.w3.org/2001/XMLSchema#double"))))
                                        //    tbl.Columns.Add(new DataColumn(item, typeof(float)));
                                        //else if (node.DataType != null && (node.DataType.Equals(UriFactory.Create("http://www.w3.org/2001/XMLSchema#integer"))))
                                        //    tbl.Columns.Add(new DataColumn(item, typeof(int)));
                                        //else
                                            tbl.Columns.Add(new DataColumn(item, typeof(string)));
                                        break;
                                    default:
                                        tbl.Columns.Add(item);
                                        break;
                                }
                            }
                            else
                                tbl.Columns.Add(item);
                        }
                    }
                    INode n;
                    String data = string.Empty;
                    if (result.TryGetValue(item, out n))
                    {
                        if (n != null)
                        {
                            switch (n.NodeType)
                            {
                                case NodeType.Uri:
                                    data = ((IUriNode)n).Uri.AbsoluteUri;
                                    break;
                                case NodeType.Blank:
                                    data = ((IBlankNode)n).InternalID;
                                    break;
                                case NodeType.Literal:
                                    //You may want to inspect the DataType and Language properties and generate
                                    //a different string here
                                    ILiteralNode node = ((ILiteralNode)n);
                                    if (!((ILiteralNode)n).Value.Equals("NAN"))
                                        data = ((ILiteralNode)n).Value;
                                    else
                                        data = "0";
                                    break;
                            }
                        }
                    }
                    vals.Add(data);
                }
                tbl.Rows.Add(vals.ToArray());
                rowNr++;
            }
            return tbl;
        }

        public static List<HierarchyStep> GetPathSteps(List<HierarchyStep> steps, string matViewLevelNum, string testViewLevelNum)
        {
            List<HierarchyStep> retSteps = null;
            List<HierarchyStep> triplesWithGrouping = steps.Where(c => c.ParentLevel.Contains(testViewLevelNum)).ToList(); //// all triples with grouping variable
            List<HierarchyStep> remainingQueryTriples = steps.Except(triplesWithGrouping).ToList();

            foreach (HierarchyStep trPtrnVal in triplesWithGrouping)
            {
                if (trPtrnVal.ChildLevel.Equals(matViewLevelNum))
                {
                    if (retSteps == null) retSteps = new List<HierarchyStep>();
                    retSteps.Add(trPtrnVal);
                }
                else  //// need to find other steps related to this one
                {
                    //// find all steps with parentNode equal to this step's child node
                    ////next portion to search for 
                    List<HierarchyStep> trPtrn = GetPathSteps(remainingQueryTriples, matViewLevelNum, trPtrnVal.ChildLevel);
                    if (trPtrn != null && trPtrn.Count > 0)
                    {
                        if (retSteps == null) retSteps = new List<HierarchyStep>();
                        retSteps.Add(trPtrnVal);
                        retSteps.AddRange(trPtrn);
                    }
                }
            }

            return retSteps;
        }
        #endregion
    }

    internal class SparqlCommands
    {
        public static string forwardPred = "http://purl.org/rdfmav#forward";
        public static string observationPred = "http://purl.org/linked-data/cube#Observation";

        public static string CmdPrefixes = @"prefix qb: <http://purl.org/linked-data/cube#> 
prefix rdf: <http://www.w3.org/1999/02/22-rdf-syntax-ns#> 
prefix rdfs: <http://www.w3.org/2000/01/rdf-schema#> 
prefix qb4o: <http://purl.org/olap#> 
prefix skos: <http://www.w3.org/2004/02/skos/core#> 
prefix xsd:   <http://www.w3.org/2001/XMLSchema#> 
prefix ex: <http://purl.org/rdfmav#> ";

        public static string CmdDimensionNames = @"
#Dimensions
SELECT ?s ?dim 
FROM NAMED <{0}>
WHERE
{{ 
  GRAPH <{0}> {{
    ?s qb:component ?obj .
    ?obj qb:dimension ?dim . 
  }}
}}";

        public static string CmdDimensionHierarchy = @"
#Hierarchies
select ?s ?hier
FROM NAMED <{0}>
WHERE{{
  GRAPH <{0}> {{
    ?s qb4o:hasHierarchy ?hier .
    FILTER(?s=<{1}>)
  }}
}}";

        public static string CmdHierarchySteps = @"
#steps in hierarchies
select ?s ?pl ?cl ?pred ?dir ?mand ?hier ?part
FROM NAMED <{0}>
WHERE {{
  GRAPH <{0}> {{
    ?s a qb4o:HierarchyStep .
    ?s qb4o:parentLevel ?pl .
    ?s qb4o:childLevel ?cl .
    ?s rdf:predicate ?pred .
    ?s ex:direction ?dir .
    ?pl qb4o:hierarchyComponent ?hier .
    OPTIONAL {{
        ?s ex:isOptional ?mand . 
    }} 
    OPTIONAL {{
        ?s ex:isPartial ?part .
    }}
    FILTER(?hier=<{1}>)
  }}
}}

";
        public static string CmdHierarchyLevels = @"
#hierarchy levels
SELECT ?dim ?hier ?lvl ?lvlName ?viewable
FROM NAMED <{0}>
WHERE
{{
  GRAPH <{0}> {{
    ?hier qb4o:inDimension ?dim ;
      qb4o:hasLevel ?lvl . 
    ?lvlName a qb4o:LevelInHierarchy ;
      qb4o:levelComponent ?lvl ;
      qb4o:hierarchyComponent ?hier .
    OPTIONAL {{
        ?lvlName ex:isViewable ?viewable .
    }}
    FILTER(?hier=<{1}> && ?dim=<{2}>)  . 
  }}
}}";
        
        public static string CmdObservationDatatype = @"
#Observation datatype
SELECT DISTINCT ?type
FROM NAMED <{0}>
WHERE
{{
  GRAPH <{0}> {{
    ?obs  rdf:type qb:DataStructureDefinition .
    ?obs rdfs:Datatype ?type . 
  }}
}}";

        /// <summary>
        /// http://example.com/ex#
        /// </summary>
        public static string CmdMeasuresInfo = @"
#Measure info
prefix rdf: <http://www.w3.org/1999/02/22-rdf-syntax-ns#> 
prefix qb: <http://purl.org/linked-data/cube#> 
prefix ex: <http://purl.org/rdfmav#> 
prefix qb4o: <http://purl.org/olap#> 

SELECT ?msrName ?agrFnct ?prdcName ?direc 
FROM NAMED <{0}>
WHERE {{
  GRAPH <{0}> {{
    ?s qb:component  ?comp .
    ?comp qb:measure ?msrName .
    ?comp qb4o:AggregateFuction ?agrFnct .
    OPTIONAL {{
      ?comp rdf:predicate ?prdcName .
      ?comp ex:direction ?direc .
    }}
  }}
}}";
    }
}
